from __future__ import absolute_import
import os
import yaml
import time
import logging
import random
import math
import uuid
import subprocess
import pytest
import sys

import kvlayer
import yakonfig

from dblogger import DatabaseLogHandler, DBLoggerQuery
from dblogger.query import complete_zulu_timestamp

config_path = os.path.join(os.path.dirname(__file__))

@pytest.fixture(scope='function')
def client(request, namespace_string, redis_address):
    config = dict(
        namespace = namespace_string,
        storage_type = 'redis',
        app_name = 'dbltest',
        storage_addresses = [redis_address],
        )
    print config
    yakonfig.set_global_config(dict(kvlayer=config))
    client = kvlayer.client()

    client.setup_namespace(
        dict(existing_table_1=2,
             existing_table_2=2))

    def cleanup():
        client.delete_namespace()
    request.addfinalizer(cleanup)

    return client

def test_basic(client):
    logger = logging.getLogger('test_logger')
    logger.setLevel(logging.DEBUG)
    dbhandler = DatabaseLogHandler(client)
    logger.addHandler(dbhandler)
    try:
        for i in xrange(10):
            logger.warn("test %d" % i)

        query = DBLoggerQuery(client)
        i = 0
        for key, record in query.filter():
            assert record.message == "test %d" % i
            i += 1
    finally:
        logger.removeHandler(dbhandler)

def test_log_exception_info(client):
    logger = logging.getLogger('test_logger')
    logger.setLevel(logging.DEBUG)
    dbhandler = DatabaseLogHandler(client)
    logger.addHandler(dbhandler)
    try:
        try:
            raise Exception('hello!')
        except:
            logger.warn("test with exception", exc_info=True)

        query = DBLoggerQuery(client)
        for key, record in query.filter():
            assert record.message
    finally:
        logger.removeHandler(dbhandler)

def test_configuration(client):
    """create the logger from the configuration, not the client object"""
    logger = logging.getLogger('test_logger')
    logger.setLevel(logging.DEBUG)
    dbhandler = DatabaseLogHandler(storage_config=client._config)
    logger.addHandler(dbhandler)
    try:
        messages = ['test {0}'.format(i) for i in xrange(10)]
        for m in messages: logger.warn(m)

        query = DBLoggerQuery(client)
        responses = [record.message for key, record in query.filter()]
        assert responses == messages
    finally:
        logger.removeHandler(dbhandler)

def test_preserve_existing_tables(client):
    logger = logging.getLogger('test_logger')
    logger.setLevel(logging.DEBUG)
    dbhandler = DatabaseLogHandler(client)
    logger.addHandler(dbhandler)

    try:
        for i in xrange(10):
            logger.warn("test %d" % i)

        key = (uuid.uuid4(), uuid.uuid4())
        val = 'data'
        client.put('existing_table_1', (key, val))
    
        assert list(client.get('existing_table_1', key))[0][1] == val
    finally:
        logger.removeHandler(dbhandler)

def test_ordering(client):
    dbhandler = DatabaseLogHandler(client)

    for i in xrange(10):
        xdict = dict(created=time.time() + (2 ** i), msg="test %d" % i)
        record = logging.makeLogRecord(xdict)
        dbhandler.emit(record)

    query = DBLoggerQuery(client)
    i = 0
    for key, record in query.filter():
        assert record.message == "test %d" % i
        i += 1

def test_queries_simple(client):
    dbhandler = DatabaseLogHandler(client)

    created_list = []
    for i in xrange(2, 10):
        created = time.time() + (2 ** i)
        created_list.append(created)
        xdict = dict(created=created, msg="test %d" % created)
        record = logging.makeLogRecord(xdict)
        dbhandler.emit(record)

    query = DBLoggerQuery(client)

    i = random.randint(0, len(created_list)-2)
    j = random.randint(i+2, len(created_list))
    slice = created_list[i:j]

    begin = math.floor(slice[0])
    end = math.ceil(slice[-1])
    i = 0
    for key, record in query.filter(begin=begin, end=end):
        assert record.message == "test %d" % slice[i]
        i += 1

    choice = random.choice(created_list)
    filter_str = 'test %d' % choice
    for key, record in query.filter(filter_str=filter_str):
        assert record.message == "test %d" % choice

def test_queries_simple2(client):
    dbhandler = DatabaseLogHandler(client)

    now = time.time()
    created_list = [now + 2 ** i for i in xrange(2,10)]
    messages = ['test {0}'.format(c) for c in created_list]
    for (created,msg) in zip(created_list, messages):
        xdict = dict(created=created, msg=msg)
        record = logging.makeLogRecord(xdict)
        dbhandler.emit(record)

    query = DBLoggerQuery(client)

    i = random.randint(0, len(created_list)-2)
    queries = created_list[i:]  ## no end
    expected = messages[i:]

    begin = math.floor(queries[0])
    end = None  ## no end
    response = query.filter(begin=begin, end=end)
    responses = [record.message for key, record in response]
    assert responses == expected

    (choice,expected) = random.choice(zip(created_list, messages))
    response = query.filter(filter_str=expected)
    responses = [record.message for key, record in response]
    assert responses == [expected]

def test_queries_cli(client):
    dbhandler = DatabaseLogHandler(client)

    created_list = []
    for i in xrange(2, 10):
        created = time.time() + (2 ** i)
        created_list.append(created)
        xdict = dict(created=created, msg="test %d" % created)
        record = logging.makeLogRecord(xdict)
        dbhandler.emit(record)

    child = subprocess.Popen(
        [sys.executable, '-m', 'dblogger.query',
         '--app-name', 'dbltest',
         '--namespace', client._config['namespace'],
         '--storage-type', client._config['storage_type'],
         '--storage-address', client._config['storage_addresses'][0],
        ],
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    child.wait()

    out = child.stdout.read()
    err = child.stderr.read()

    assert child.returncode == 0
    assert out

def test_complete_zulu_timestamp():
    assert complete_zulu_timestamp('1998-10') == '1998-10-01T00:00:00.000000Z'

def test_queries_cli2(client):
    dbhandler = DatabaseLogHandler(client)

    created_list = []
    for i in xrange(2, 10):
        created = time.time() + (2 ** i)
        created_list.append(created)
        xdict = dict(created=created, msg="test %d" % created)
        record = logging.makeLogRecord(xdict)
        dbhandler.emit(record)

    child = subprocess.Popen(
        [sys.executable, '-m', 'dblogger.query',
         '--app-name', 'dbltest',
         '--namespace', client._config['namespace'],
         '--storage-type', client._config['storage_type'],
         '--storage-address', client._config['storage_addresses'][0],
         '--begin', '1998-01-03T08',
        ],
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    child.wait()

    out = child.stdout.read()
    err = child.stderr.read()

    assert child.returncode == 0, err
    assert out



'''Python logging setup and database backend.

.. This software is released under an MIT/X11 open source license.
   Copyright 2013-2014 Diffeo, Inc.

.. only:: pyapi or all

    This package provides basic support tools for the standard Python
    :mod:`logging` package.  In particular, it exposes the
    :mod:`logging.config` setup language through configuration layers
    such as :mod:`yakonfig`, and it includes some basic database
    support utilities.

    Applications that use this support a ``logging`` block in their
    configuration files.  This may have any valid logging setup
    supported by the :mod:`logging.config` module.  The default
    configuration includes a formatter named ``fixed`` and a handler
    named ``console`` bound to the root logger, and a further console
    handler named ``debug`` not bound anywhere.

.. only:: all

    Command-line options
    ====================

    All programs that use :mod:`dblogger` support shared command-line
    options.  If not changed in the configuration, the defaults write
    :data:`logging.INFO` level output to the console, and
    :option:`--verbose` enables all debug-level logging.

.. only:: cliref or all

    .. program:: dblogger

    .. option:: --verbose, -v

        Include more output, increasing the ``console`` log handler's
        level one step.  This can be included multiple times.  If the
        program is more verbose than quiet and the ``console`` log
        handler is not attached to the root logger, adds it.

    .. option:: --quiet, -q

        Include less output, decreasing the ``console`` log handler's
        level one step.  This can be included multiple times.

    .. option:: --debug <logger>

        Write debug-level output for `logger` to the ``debug`` log
        handler.  This can be included multiple times.

    Recipes
    =======

    To enable debug-level logging everywhere, include a top-level
    block in the YAML configuration file:

    .. code-block:: yaml

        logging:
          root:
            level: DEBUG

    To enable debug-level logging for a specific package in the
    configuration file:

    .. code-block:: yaml

        logging:
          loggers:
            some.other.package:
              level: DEBUG

    To write debug logs globally to a file:

    .. code-block:: yaml

        logging:
          handlers:
            file:
              class: logging.FileHandler
              filename: /tmp/dblogger.log
              formatter: fixed
          root:
            handlers: [console, file]

    To write logs to a database and suppress normal console output:

    .. code-block:: yaml

        logging:
          handlers:
            dblogger:
              class: dblogger.DatabaseLogHandler
              storage_config: *kvlayer
          root:
            handlers: [dblogger]
        kvlayer: &kvlayer
          storage_type: redis
          storage_addresses: ['redis.example.com:6379']

    In this last case, :mod:`dblogger <dblogger.query>` can print out
    the saved log messages.

    For further details about what is allowed, see the Python library
    :mod:`logging.config` documentation.

.. only:: pyapi or all

    Logging tools
    =============

    .. autoclass:: DatabaseLogHandler
       :show-inheritance:

    .. autoclass:: FixedWidthFormatter
       :show-inheritance:

    Initial logging setup
    =====================

    Include :mod:`dblogger` in your :func:`yakonfig.parse_args` call
    to do all of the initial setup.  When control returns to your
    application, logging will be fully set up.

    Older programs that do not yet support this interface use:

    .. autofunction:: configure_logging

.. only:: all

    :program:`dblogger` query tool
    ==============================

    .. automodule:: dblogger.query

'''
from __future__ import absolute_import
from dblogger.configure import config_name, default_config, add_arguments, \
    runtime_keys, normalize_config
from dblogger.configure import configure_logging
from dblogger.format import FixedWidthFormatter
from dblogger.logger import DatabaseLogHandler
from dblogger.query import *



'''Query tool to retrieve :class:`dblogger.DatabaseLogHandler` results.

.. This software is released under an MIT/X11 open source license.
   Copyright 2013-2014 Diffeo, Inc.

:program:`dblogger` retrieves messages stored in a database using
:class:`dblogger.DatabaseLogHandler`.

This supports the standard :option:`--config <yakonfig --config>`,
:option:`--dump-config <yakonfig --dump-config>`,
:option:`--verbose <dblogger --verbose>`,
:option:`--quiet <dblogger --quiet>`, and
:option:`--debug <dblogger --debug>` options.

.. program:: dblogger

.. option:: --begin <time>

    Only show messages at or after `time`.  Time should be an ISO date
    string of the form ``YYYY-MM-DDTHH:MM:SS.MMMMMMZ``, or any
    left-truncated prefix thereof.

.. option:: --end <time>

    Only show messages at or before `time`.

'''
from __future__ import absolute_import
import argparse
import logging
import re
import sys
import time

import dblogger
from dblogger.format import FixedWidthFormatter
from dblogger.logger import DatabaseLogHandler
from dblogger.utils import gen_uuid
import kvlayer
import streamcorpus
import yakonfig

class DBLoggerQuery(object):
    def __init__(self, storage_client, table_name="log"):
        self.storage = storage_client
        self.table_name = table_name
        self.last_uuid = None
        storage_client.setup_namespace({ table_name : 1 })

    def build_key_range(self, uuid_start=None, uuid_end=None):
        key_start = tuple()
        key_end = tuple()

        if uuid_start:
            key_start = (uuid_start,)

        if uuid_end:
            key_end = (uuid_end,)

        return (key_start, key_end)


    def filter(self, begin=None, end=None, filter_str=None, tail=False):
        """Get log record from the database.

        begin and end must be timestamp as returned by time.time().

        filter_str() -- An dict of filters that will match agaist log record
        fields. Not Implemented yet.

        """

        uuid_begin = uuid_end = None
        if begin:
            uuid_begin = gen_uuid(begin)
        if end:
            uuid_end = gen_uuid(end)
        key_range = self.build_key_range(uuid_begin, uuid_end)

        if filter_str:
            filter_re = re.compile(filter_str)

        while True:
            for key, rec_pickle in self.storage.scan(self.table_name, key_range):

                ## what is the purpose of these three lines?
                if key[0] == self.last_uuid:
                    break
                self.last_uuid = key[0]
                ##  ^^^^^ why? ^^^^^^^^

                record = DatabaseLogHandler.deserialize(rec_pickle)
                if filter_str and not filter_re.match(record.message):
                    continue
                yield key, record

            if not tail:
                break

            time.sleep(1)
            key_range = self.build_key_range(uuid_begin=self.last_uuid)


zulu_timestamp_re = re.compile('(?P<year>\d{4})?-?(?P<month>\d{0,2})?-?(?P<day>\d{0,2})?T?(?P<hour>\d{0,2})?:?(?P<minute>\d{0,2})?:?(?P<second>\d{0,2})?.?(?P<microsecond>\d{0,6})?Z?')
#%Y-%m-%dT%H:%M:%S.%fZ'

def complete_zulu_timestamp(partial_zulu_timestamp):
    'fill out a partial zulu_timestamp string'
    match = zulu_timestamp_re.match(partial_zulu_timestamp)
    vals = map(lambda x: x and int(x) or 0, match.groups())
    if  vals[1] == 0:
        vals[1] = 1
    if  vals[2] == 0:
        vals[2] = 1
    return '%d-%02d-%02dT%02d:%02d:%02d.%06dZ' % tuple(vals)


def main():
    parser = argparse.ArgumentParser(
        description='retrieve log messages from kvlayer')
    parser.add_argument(
        '--begin', default=None, 
        help='YYYY-MM-DDTHH:MM:SS.MMMMMMZ in UTC can be truncated at any depth')
    parser.add_argument(
        '--end', default=None, 
        help='YYYY-MM-DDTHH:MM:SS.MMMMMMZ in UTC can be truncated at any depth')
    parser.add_argument('--past', type=int, default=60, metavar='N',
                        help='show the most recent N seconds of logs, default N=60. '
                        'If negative, then scan from earliest moment *upto* '
                        'N seconds ago.')
    parser.add_argument('--clear', action='store_true', default=False,
                        help='delete all messages in scan')
    parser.add_argument('-y', '--yes', default=False, action='store_true',
                        dest='assume_yes',
                        help='assume "yes" and require no input for '
                        'confirmation questions.')
    args = yakonfig.parse_args(parser, [yakonfig, kvlayer, dblogger])

    if args.begin:
        args.begin = streamcorpus.make_stream_time(
            complete_zulu_timestamp(args.begin)).epoch_ticks
    elif args.past > 0:
        args.begin = time.time() - args.past
    else:
        args.begin = None

    if args.end:
        args.end   = streamcorpus.make_stream_time(
            complete_zulu_timestamp(args.end)).epoch_ticks
    elif args.past < 0:
        args.end = time.time() + args.past
    else:
        args.end = None

    if args.clear:
        namespace = yakonfig.get_global_config('kvlayer')['namespace']
        if not args.assume_yes:
            response = raw_input('Delete all logs between {!r} and {!r} in {!r}?  Enter namespace: '
                                 .format(args.begin, args.end, namespace))
            if response != namespace:
                sys.stdout.write('not deleting anything\n')
                return
        sys.stdout.write('deleting logs between {!r} and {!r} in namespace {!r}\n'
                          .format(args.begin, args.end, namespace))

    client = kvlayer.client()
    query = DBLoggerQuery(client)

    ## rather than specify the format string in a second place, we get
    ## the configured StreamHandler and use its `format` method
    logger = logging.getLogger()
    ch = None
    for handler in logger.handlers:
        if isinstance(handler, logging.StreamHandler):
            ch = handler
    assert ch, 'must have a StreamHandler configured in order to use dblogger command-line'

    count = 0
    for key, record in query.filter(args.begin, args.end):
        print ch.format(record)
        if args.clear:
            client.delete(query.table_name, key)
        count += 1
    if count == 0:
        print 'no log records found'
    else:
        print 'returned %d log records' % count
        if args.clear:
            print 'and deleted all of them.'

if __name__ == '__main__':
    main()



"""tests for dblogger.configure"""

import logging
from StringIO import StringIO

import yaml

from dblogger import configure_logging

def test_default_config(capsys):
    configure_logging({})
    logger = logging.getLogger('dblogger.test_global')
    logger.critical('test')
    (out,err) = capsys.readouterr()
    print out
    # this comes from the default format string...we can't test everything,
    # but we can at least check things like
    assert err[34:46] == 'test_configu'
    assert err[52:] == 'CRITICAL test\n'

def test_change_log_level_only(capsys):
    config = """
    logging:
        version: 1
        root:
            level: DEBUG
    """
    config = yaml.load(StringIO(config))
    configure_logging(config)
    logger = logging.getLogger('dblogger.test_global')
    logger.critical('test')
    logger.debug('test 2')
    (out,err) = capsys.readouterr()
    print out
    assert err[34:46] == 'test_configu'
    assert err[52:66] == 'CRITICAL test\n'
    assert err[100:112] == 'test_configu'
    assert err[118:] == 'DEBUG    test 2\n'

def test_toplevel_config(capsys):
    config = """
    logging:
        version: 1
        formatters:
            toplevelf:
                format: 'toplevel %(levelname)s %(message)s'
        handlers:
            toplevel:
                class: logging.StreamHandler
                formatter: toplevelf
        root:
            handlers: [toplevel]
    """
    config = yaml.load(StringIO(config))
    configure_logging(config)
    logger = logging.getLogger('dblogger.test_global')
    logger.critical('test')
    (out,err) = capsys.readouterr()
    print out
    assert err == 'toplevel CRITICAL test\n'

def test_inline_config(capsys):
    config = """
    version: 1
    formatters:
        inlinef:
            format: 'inline %(levelname)s %(message)s'
    handlers:
        inline:
            class: logging.StreamHandler
            formatter: inlinef
    root:
        handlers: [inline]
    """
    config = yaml.load(StringIO(config))
    configure_logging(config)
    logger = logging.getLogger('dblogger.test_global')
    logger.critical('test')
    (out,err) = capsys.readouterr()
    print out
    assert err == 'inline CRITICAL test\n'





import time
import pytest
import kvlayer
import logging
import multiprocessing

import yakonfig

from dblogger.tests.test_dblogger import client ## fixture that cleans up

from dblogger import DatabaseLogHandler, DBLoggerQuery


ch = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s pid=%(process)d %(filename)s:%(lineno)d %(levelname)s: %(message)s')
ch.setLevel('DEBUG')
ch.setFormatter(formatter)
logger = logging.getLogger()
logger.addHandler(ch)

def worker(config, num_records):
    yakonfig.set_global_config(dict(kvlayer=config))
    client = kvlayer.client()
    dbhandler = DatabaseLogHandler(client)
    logger = logging.getLogger('foo')
    logger.addHandler(dbhandler)
    for i in xrange(num_records):
        logger.critical('a message: %d', i)
    logger.critical('finished')

def cleanup(results):
    for res in results:
        try:
            res.get(0)
        except multiprocessing.TimeoutError:
            continue
        except Exception, exc:
            logger.critical('trapped child exception', exc_info=True)
            raise
        else:
            ## if it gets here, slot should always be finished
            assert res.ready()
            results.remove(res)

    return len(results)

@pytest.mark.performance
def test_queries_throughput(client):
    num_tasks = 10
    num_records = 100
    num_children = 10
    pool = multiprocessing.Pool(num_children, maxtasksperchild=1)
    results = []
    for x in xrange(num_tasks):
        results.append(
            pool.apply_async(worker, args=(client._config, num_records)))

    query = DBLoggerQuery(client)

    timeout = 60
    start_time = time.time()
    while time.time() - start_time < timeout:
        if cleanup(results) == 0:
            logger.critical('finished')
            break

    elapsed = time.time() - start_time
    expected_count = num_records * num_tasks + num_tasks
    logger.critical('finished logging %d messages in %.1f sec --> %.1f per sec',
                    expected_count, elapsed, float(expected_count) / elapsed)

    count = 0
    for key, record in query.filter():
        logger.critical( record )
        assert record.message
        count += 1
    
    assert count == expected_count




import random
import struct
import time

from uuid import UUID


def gen_uuid(timestamp=None, sequence=None):
    if timestamp is None:
        timestamp = time.time()
    if sequence is None:
        sequence = random.getrandbits(32)

    # 128 bit fake-uuid is 64 bits time, 64 bits random number.
    # High order bits first to keep ordering.
    bytes = struct.pack('>qLL', long(timestamp * 1024), sequence,
                        random.getrandbits(32))

    return UUID(bytes=bytes)


def random_slice(items):
    start = random.randint(0, len(items))
    end = random.randint(0, len(items))
    return items[start:end+1]


def datetime_to_time(datetime):
    return time.mktime(datetime.timetuple())



'''Formatters for :mod:`logging`.

.. This software is released under an MIT/X11 open source license.
   Copyright 2013-2014 Diffeo, Inc.

'''

import logging

class FixedWidthFormatter(logging.Formatter):
    '''Formats log messages in fixed columns.

    This adds format string properties
    ``%(fixed_width_filename_lineno)s`` containing the file name and
    line number in a 17-character-wide field, and
    ``%(fixed_width_levelname`` containing the log level padded out to
    an 8-character-wide field.

    '''
    filename_width = 17
    levelname_width = 8

    def format(self, record):
        max_filename_width = self.filename_width - 3 - len(str(record.lineno))
        filename = record.filename
        if len(record.filename) > max_filename_width:
            filename = record.filename[:max_filename_width]
        a = "%s:%s" % (filename, record.lineno)
        record.fixed_width_filename_lineno = a.ljust(self.filename_width)
        levelname = record.levelname
        levelname_padding = ' ' * (self.levelname_width - len(levelname))
        record.fixed_width_levelname = levelname + levelname_padding
        return super(FixedWidthFormatter, self).format(record)





''':mod:`logging` handler that stores log messages in a database.

.. This software is released under an MIT/X11 open source license.
   Copyright 2013-2014 Diffeo, Inc.

'''
from __future__ import absolute_import

import time
import logging
import cPickle as pickle
import sys
import traceback

try:
    from tblib import pickling_support
    pickling_support.install()  # register traceback smarts with pickle
except Exception, exc:
    ## tblib does not work in python2.6...
    ## log something?
    pass

from dblogger.utils import gen_uuid
import kvlayer
import yakonfig


class DatabaseLogHandler(logging.Handler):
    '''Log handler that stores log messages in a database.

    This uses :mod:`kvlayer` to store the actual log messages.
    When the log handler is created, the caller needs to pass
    in either the :mod:`kvlayer` configuration or the actual
    :class:`kvlayer.AbstractStorage` object.  The constructor
    also accepts a virtual table name, defaulting to ``log``.

    If a global YAML file is used to configure the application, then
    YAML reference syntax can be used to share this handler's
    configuration with the global kvlayer configuration.

    .. code-block:: yaml

        kvlayer: &kvlayer
          storage_type: redis
          storage_addresses: [ 'redis.example.com:6379' ]
        logging:
          handlers:
            db:
              class: dblogger.DatabaseLogHandler
              storage_config: *kvlayer

    Log messages are stored in a table with a single UUID key, where
    the high-order bits of the UUID are in order by time.  The actual
    table values are serialized JSON representations of the log
    records.

    This log handlers adds a format string property ``%(humantime)s``
    with a time in a fixed format, and ``%(exc_text)s`` with the
    formatted traceback from an exception.  These properties are also
    included in the JSON stored in the database..

    .. automethod:: __init__

    '''
    def __init__(self, storage_client=None, table_name="log",
                 storage_config=None):
        """Create a new database log handler.

        You must either pass in ``storage_client``, an actual kvlayer
        client object, or ``storage_config``, a dictionary which will
        be passed to ``kvlayer.client()``.  Log messages
        will be stored in the table ``table_name``.

        :param storage_client: existing storage client
        :type storage_client: :class:`kvlayer.AbstractStorage`
        :param str table_name: virtual table name
        :param dict storage_config: configuration for new storage client

        """
        super(DatabaseLogHandler, self).__init__()

        if storage_client is None:
            if storage_config is None:
                raise RuntimeError('must pass either storage_client or '
                                   'storage_config')
            with yakonfig.defaulted_config(
                    [kvlayer], config=dict(kvlayer=storage_config)):
                storage_client = kvlayer.client()

        self.storage = storage_client
        self.table_name = table_name
        storage_client.setup_namespace({table_name: 1})
        self.sequence_number = 0

    def formatDBTime(self, record):
        record.humantime = time.strftime('%Y-%m-%dT%H:%M:%S-%Z',
                                         time.localtime(record.created))

    @classmethod
    def deserialize(cls, rec_pickle):
        try:
            xdict = pickle.loads(rec_pickle)

        except Exception:
            xdict = {'msg': 'warning!!!! failed to unpickle: %r' % rec_pickle}

        return logging.makeLogRecord(xdict)

    def emit(self, record):
        '''
        handle a record by formatting parts of it, and pushing it into
        storage.
        '''
        self.format(record)
        self.formatDBTime(record)

        failure = []
        if record.args:
            try:
                # cannot serialize arbitrary args, because they might not be
                # picklable, so do the string now
                record.msg = record.msg % record.args
                record.args = None
            except Exception, exc:
                failure.append('failed to run string formatting on '
                               'provided args')
                failure.append(traceback.format_exc(exc))
                failure.append('record.msg = %r' % record.msg)
                failure.append('record.args = %r' % (record.args,))
                failure.append('logging failed so shutting down entire '
                               'process')

        if record.exc_info:
            record.exc_text = logging._defaultFormatter.formatException(
                record.exc_info)
        else:
            record.exc_text = ''

        # NB: This is safe because emit() is called from handler() under
        # self.lock
        new_uuid = gen_uuid(record.created, self.sequence_number)
        self.sequence_number += 1

        try:
            dbrec = pickle.dumps(record.__dict__,
                                 protocol=pickle.HIGHEST_PROTOCOL)
        except Exception, exc:
            failure.append('failed to dump log record, will shutdown')
            failure.append(traceback.format_exc(exc))
            failure.append('failed to pickle the __dict__ on: record=%r'
                           % record)
            failure.append('failed to pickle: record.__dict__=%r'
                           % record.__dict__)
            failure.append('logging failed so shutting down entire process')

        if failure:
            dbrec = '\n'.join(failure)

        # send it to the DB... especially if it is a failure
        self.storage.put(self.table_name, ((new_uuid,), dbrec))

        if failure:
            # shutdown the process when logging fails
            sys.exit(dbrec)



"""Globally configure logging from a dictionary.

.. This software is released under an MIT/X11 open source license.
   Copyright 2014 Diffeo Inc.

Purpose
=======

Call :func:`dblogger.configure_logging`` with a configuration
dictionary, such as the global configuration given to :mod:`yakonfig`.
This will perform the logging configuration described there, and set
reasonable defaults if there is no configuration.

If the dictionary passed to :func:`configure_logging` contains a key
``logging``, that key's value is used as the logging configuration;
otherwise if the dictionary looks like logging configuration itself,
it is used directly; otherwise only default settings are used.

Module Contents
===============

"""

from __future__ import absolute_import

import logging.config

# yakonfig configuration metadata
config_name = 'logging'
default_config = {
    'version': 1,
    'disable_existing_loggers': False,
    'verbose': 0,
    'quiet': 0,
    'debug': [],
    'formatters': {
        'fixed': {
            '()': 'dblogger.FixedWidthFormatter',
            'format': ('%(asctime)-23s pid=%(process)-5d '
                       '%(fixed_width_filename_lineno)s '
                       '%(fixed_width_levelname)s %(message)s'),
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'fixed',
            'level': 'INFO',
        },
        'debug': {
            'class': 'logging.StreamHandler',
            'formatter': 'fixed',
            'level': 'DEBUG',
        },
    },
    'root': {
        'handlers': ['console'],
        'level': 'NOTSET',
    },
    'loggers': {
    },
}
def add_arguments(parser):
    parser.add_argument('--verbose', '-v', action='count',
                        help='produce more output')
    parser.add_argument('--quiet', '-q', action='count',
                        help='produce less output')
    parser.add_argument('--debug', action='append', metavar='LOGGER',
                        help='enable debug logging for specific modules')
runtime_keys = {
    'verbose': 'verbose',
    'quiet': 'quiet',
    'debug': 'debug',
}
def normalize_config(config):
    # Adjust console logging level
    verbosity = config.get('verbose', 0) - config.get('quiet', 0)
    if 'handlers' in config:
        if 'console' in config['handlers']:
            level = config['handlers']['console'].get('level', logging.INFO)
            if not isinstance(level, int):
                level = logging.getLevelName(level)
            level -= verbosity * 10
            # adjust to be in bounds
            level = min(level, logging.CRITICAL)
            level = max(level, logging.NOTSET)
            # bring back to a string
            try:
                level = logging.getLevelName(level)
            except KeyError, e:
                pass
            # stash
            config['handlers']['console']['level'] = level

    # Re-bind console logger if required
    if verbosity > 0:
        config.setdefault('root', {})
        config['root'].setdefault('handlers', [])
        if 'console' not in config['root']['handlers']:
            config['root']['handlers'].append('console')

    # Enable debug logging
    config.setdefault('loggers', {})
    for logger in config.get('debug', []):
        config['loggers'].setdefault(logger, {})
        config['loggers'][logger].setdefault('handlers', [])
        config['loggers'][logger]['handlers'].append('debug')

    # We've used these options and integrated them into other things,
    # so reset them
    for option in ('verbose', 'quiet', 'debug'):
        config[option] = default_config[option]

    # Fire everything off
    logging.config.dictConfig(config)

def configure_logging(config):
    """One-time global logging configuration.

    Set up logging as described in ``config``.  ``config`` should be a
    top-level configuration dictionary with a key ``logging``, and the
    value of that key is used as a configuration dictionary for
    :mod:`logging.config`.  If there is no ``logging`` key but there
    is a ``version: 1`` key/value, ``config`` is used directly as the
    configuration.  Otherwise a minimal default configuration is used.

    If the configuration does not define any handlers, then a default
    console log handler will be created and bound to the root logger.
    This will use a formatter named ``fixed``, defining it if
    necessary.

    :param dict config: :mod:`logging.config` setup dictionary

    .. deprecated:: 0.4.0
        Pass :mod:`dblogger` to :func:`yakonfig.parse_args` instead.

    """
    # find the actual logging config dictionary
    if 'logging' in config:
        config = config['logging']
        config.setdefault('version', 1)
    elif config.get('version') == 1:
        config = config
    else:
        config = { 'version': 1 }

    # create default handler if required
    if len(config.setdefault('handlers', {})) == 0:
        config.setdefault('formatters', {})
        if 'fixed' not in config['formatters']:
            config['formatters']['fixed'] = {
                '()': 'dblogger.FixedWidthFormatter',
                'format': ('%(asctime)-23s pid=%(process)-5d '
                           '%(fixed_width_filename_lineno)s '
                           '%(fixed_width_levelname)s %(message)s'),
            }
        config['handlers']['console'] = {
            'class': 'logging.StreamHandler',
            'formatter': 'fixed',
        }
        config.setdefault('root', {})
        config['root']['handlers'] = ['console']
    
    # also, we must set this magic flag, or any logger created at the
    # module level will stop working
    config['disable_existing_loggers'] = False

    logging.config.dictConfig(config)



#!/usr/bin/env python

import os
import sys
import fnmatch
import subprocess

## prepare to run PyTest as a command
from distutils.core import Command

from setuptools import setup, find_packages

from version import get_git_version
VERSION, SOURCE_LABEL = get_git_version()
PROJECT = 'dblogger'
URL = 'https://github.com/diffeo/dblogger'
AUTHOR = 'Diffeo, Inc.'
AUTHOR_EMAIL = 'support@diffeo.com'
DESC = 'DB-backed python logging.Handler subclass that uses kvlayer, and provides command-line tools.'
LICENSE = 'MIT/X11 license http://opensource.org/licenses/MIT'


def read_file(file_name):
    file_path = os.path.join(
        os.path.dirname(__file__),
        file_name
    )
    return open(file_path).read()


def recursive_glob(treeroot, pattern):
    results = []
    for base, dirs, files in os.walk(treeroot):
        goodfiles = fnmatch.filter(files, pattern)
        results.extend(os.path.join(base, f) for f in goodfiles)
    return results


def _myinstall(pkgspec):
    setup(
        script_args = ['-q', 'easy_install', '-v', pkgspec],
        script_name = 'easy_install'
    )


class InstallTestDependencies(Command):
    '''install test dependencies'''

    description = 'installs all dependencies required to run all tests'

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def easy_install(self, packages):
        cmd = ['easy_install']
        if packages:
            cmd.extend(packages)
            errno = subprocess.call(cmd)
            if errno:
                raise SystemExit(errno)

    def run(self):
        if self.distribution.install_requires:
            self.easy_install(self.distribution.install_requires)
        if self.distribution.tests_require:
            self.easy_install(self.distribution.tests_require)


class PyTest(Command):
    '''run py.test'''

    description = 'runs py.test to execute all tests'

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        if self.distribution.install_requires:
            for ir in self.distribution.install_requires:
                _myinstall(ir)
        if self.distribution.tests_require:
            for ir in self.distribution.tests_require:
                _myinstall(ir)

        # reload sys.path for any new libraries installed
        import site
        site.main()
        print sys.path
        # use pytest to run tests
        pytest = __import__('pytest')
        if pytest.main(['-n', '8', '-s', 'dblogger', '--runperf', '--runslow']):
            sys.exit(1)

setup(
    name=PROJECT,
    version=VERSION,
    source_label=SOURCE_LABEL,
    description=DESC,
    license=LICENSE,
    long_description=read_file('README.rst'),
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=find_packages(),
    cmdclass={'test': PyTest,
              'install_test': InstallTestDependencies},
    # We can select proper classifiers later
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',  ## MIT/X11 license http://opensource.org/licenses/MIT
    ],
    tests_require=[
        'pytest',
        'ipdb',
        'pytest-cov',
        'pytest-xdist',
        'pytest-timeout',
        'pytest-incremental',
        'pytest-capturelog',
        'epydoc',
        'pytest-diffeo',
    ],
    install_requires=[
        'tblib',  ## to pickle tracebacks
        'yakonfig >= 0.4.1',
        'kvlayer >= 0.4.0',
        'streamcorpus',
        'python-dateutil',
    ],
    entry_points={
        'console_scripts': [
            'dblogger = dblogger.query:main',
        ]
    },
)
